package com.chelvc.framework.export;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.time.Duration;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;
import java.util.stream.Collectors;
import javax.annotation.PostConstruct;
import javax.servlet.http.HttpServletResponse;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.ThreadContextHolder;
import com.chelvc.framework.base.util.HttpUtils;
import com.chelvc.framework.common.util.AssertUtils;
import com.chelvc.framework.common.util.ExcelUtils;
import com.chelvc.framework.common.util.FileUtils;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.common.util.StringUtils;
import com.chelvc.framework.common.util.ThreadUtils;
import com.chelvc.framework.export.config.ExportProperties;
import com.chelvc.framework.redis.context.RedisContextHolder;
import com.google.common.collect.Lists;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.streaming.SXSSFWorkbook;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Component;

/**
 * 数据导出默认实现
 *
 * @author Woody
 * @date 2024/5/9
 */
@Slf4j
@Component
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class DefaultExportHandler implements ExportHandler {
    private final ExportProperties properties;
    private final RedisTemplate<String, Object> redisTemplate;

    /**
     * 清理导出文件
     */
    private void clear() {
        int batch = 100;
        List<File> files = Lists.newArrayListWithCapacity(batch);
        FileUtils.files(this.properties.getDirectory(), file -> {
            files.add(file);
            if (files.size() == batch) {
                this.clear(files);
                files.clear();
            }
        });
        if (ObjectUtils.notEmpty(files)) {
            this.clear(files);
        }
    }

    /**
     * 清理导出文件
     *
     * @param files 文件列表
     */
    private void clear(List<File> files) {
        List<String> keys = files.stream().map(File::getName).map(this::key).collect(Collectors.toList());
        List<?> values = this.redisTemplate.opsForValue().multiGet(keys);
        if (ObjectUtils.notEmpty(values)) {
            for (int i = 0; i < values.size(); i++) {
                if (Objects.isNull(values.get(i))) {
                    FileUtils.delete(files.get(i));
                }
            }
        }
    }

    /**
     * 构建任务标识
     *
     * @param id 任务ID
     * @return 任务标识
     */
    private String key(Serializable id) {
        return "exporting:" + id;
    }

    /**
     * 获取任务状态
     *
     * @param id 任务ID
     * @return 任务状态信息
     */
    private Pair<String, Boolean> pair(Serializable id) {
        String value = (String) this.redisTemplate.opsForValue().get(this.key(id));
        if (StringUtils.isEmpty(value)) {
            return null;
        }
        String[] splits = value.split(":");
        return Pair.of(splits[0], Boolean.valueOf(splits[1]));
    }

    /**
     * 更新任务状态
     *
     * @param id       任务ID
     * @param pair     任务状态信息
     * @param duration 有效时间
     */
    private void pair(Serializable id, Pair<String, Boolean> pair, Duration duration) {
        String value = pair.getLeft() + ":" + pair.getRight();
        this.redisTemplate.opsForValue().set(this.key(id), value, duration);
    }

    @PostConstruct
    public void initialize() {
        ThreadUtils.run(() -> {
            while (!Thread.currentThread().isInterrupted()) {
                try {
                    this.clear();
                } catch (Exception e) {
                    log.error("Export file clear failed", e);
                }
                ThreadUtils.sleep(60 * 1000);
            }
        });
    }

    @Override
    public boolean isCompleted(long id) {
        Boolean completed = ObjectUtils.ifNull(this.pair(id), Pair::getRight);
        return Boolean.TRUE.equals(completed);
    }

    @Override
    public void download(long id, @NonNull HttpServletResponse response) throws IOException {
        Pair<String, Boolean> pair = this.pair(id);
        AssertUtils.available(pair, () -> ApplicationContextHolder.getMessage("Export.Expired"));
        boolean completed = Boolean.TRUE.equals(ObjectUtils.ifNull(pair, Pair::getRight));
        AssertUtils.available(completed, () -> ApplicationContextHolder.getMessage("Export.Uncompleted"));
        String filename = ObjectUtils.ifNull(pair, Pair::getLeft);
        HttpUtils.write(response, new File(this.properties.getDirectory(), String.valueOf(id)), filename);
    }

    @Override
    public <T> long export(@NonNull String filename, @NonNull BiConsumer<Long, Integer> listener,
                           @NonNull BiFunction<Integer, Integer, Collection<T>> provider,
                           @NonNull ExcelUtils.Writer<T> writer, @NonNull String... titles) {
        // 生成导出任务ID
        long id = RedisContextHolder.identity();

        // 初始化数据导出状态
        this.pair(id, Pair.of(filename, false), Duration.ofDays(1));

        // 异步执行导出任务
        ThreadContextHolder.execute(() -> {
            // 将数据写入工作薄
            Workbook workbook = new SXSSFWorkbook();
            int volume = ExcelUtils.write(workbook, provider, writer, titles);

            // 保存导出文件
            File file = new File(this.properties.getDirectory(), String.valueOf(id));
            try (OutputStream output = new FileOutputStream(file)) {
                workbook.write(output);
                workbook.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }

            // 更新数据导出任务状态
            this.pair(id, Pair.of(filename, true), Duration.ofSeconds(this.properties.getExpiration()));

            // 数据导出完成监听回调
            listener.accept(id, volume);
        });
        return id;
    }
}
